/******************************************************************************/
/*                                                                            */
/*    Copyright (c) 1990-2016, KAIST                                          */
/*    All rights reserved.                                                    */
/*                                                                            */
/*    Redistribution and use in source and binary forms, with or without      */
/*    modification, are permitted provided that the following conditions      */
/*    are met:                                                                */
/*                                                                            */
/*    1. Redistributions of source code must retain the above copyright       */
/*       notice, this list of conditions and the following disclaimer.        */
/*                                                                            */
/*    2. Redistributions in binary form must reproduce the above copyright    */
/*       notice, this list of conditions and the following disclaimer in      */
/*       the documentation and/or other materials provided with the           */
/*       distribution.                                                        */
/*                                                                            */
/*    3. Neither the name of the copyright holder nor the names of its        */
/*       contributors may be used to endorse or promote products derived      */
/*       from this software without specific prior written permission.        */
/*                                                                            */
/*    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS     */
/*    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       */
/*    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       */
/*    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE          */
/*    COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,    */
/*    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    */
/*    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        */
/*    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        */
/*    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      */
/*    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       */
/*    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         */
/*    POSSIBILITY OF SUCH DAMAGE.                                             */
/*                                                                            */
/******************************************************************************/
/******************************************************************************/
/*                                                                            */
/*    ODYSSEUS/OOSQL DB-IR-Spatial Tightly-Integrated DBMS                    */
/*    Version 5.0                                                             */
/*                                                                            */
/*    Developed by Professor Kyu-Young Whang et al.                           */
/*                                                                            */
/*    Advanced Information Technology Research Center (AITrc)                 */
/*    Korea Advanced Institute of Science and Technology (KAIST)              */
/*                                                                            */
/*    e-mail: odysseus.oosql@gmail.com                                        */
/*                                                                            */
/*    Bibliography:                                                           */
/*    [1] Whang, K., Lee, J., Lee, M., Han, W., Kim, M., and Kim, J., "DB-IR  */
/*        Integration Using Tight-Coupling in the Odysseus DBMS," World Wide  */
/*        Web, Vol. 18, No. 3, pp. 491-520, May 2015.                         */
/*    [2] Whang, K., Lee, M., Lee, J., Kim, M., and Han, W., "Odysseus: a     */
/*        High-Performance ORDBMS Tightly-Coupled with IR Features," In Proc. */
/*        IEEE 21st Int'l Conf. on Data Engineering (ICDE), pp. 1104-1105     */
/*        (demo), Tokyo, Japan, April 5-8, 2005. This paper received the Best */
/*        Demonstration Award.                                                */
/*    [3] Whang, K., Park, B., Han, W., and Lee, Y., "An Inverted Index       */
/*        Storage Structure Using Subindexes and Large Objects for Tight      */
/*        Coupling of Information Retrieval with Database Management          */
/*        Systems," U.S. Patent No.6,349,308 (2002) (Appl. No. 09/250,487     */
/*        (1999)).                                                            */
/*    [4] Whang, K., Lee, J., Kim, M., Lee, M., Lee, K., Han, W., and Kim,    */
/*        J., "Tightly-Coupled Spatial Database Features in the               */
/*        Odysseus/OpenGIS DBMS for High-Performance," GeoInformatica,        */
/*        Vol. 14, No. 4, pp. 425-446, Oct. 2010.                             */
/*    [5] Whang, K., Lee, J., Kim, M., Lee, M., and Lee, K., "Odysseus: a     */
/*        High-Performance ORDBMS Tightly-Coupled with Spatial Database       */
/*        Features," In Proc. 23rd IEEE Int'l Conf. on Data Engineering       */
/*        (ICDE), pp. 1493-1494 (demo), Istanbul, Turkey, Apr. 16-20, 2007.   */
/*                                                                            */
/******************************************************************************/

#ifndef _OOSQL_EXTERNALPROCESSMESSAGE_H_
#define _OOSQL_EXTERNALPROCESSMESSAGE_H_

enum OOSQL_ExternalProcessMessageID {
	OOSQL_EXTPROCMSG_ID_CALL_FUNCTION,
	OOSQL_EXTPROCMSG_ID_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_EXIT,
	OOSQL_EXTPROCMSG_ID_EXIT_DONE,
	OOSQL_EXTPROCMSG_ID_SYSTEM_HANDLE,

	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_CREATESYSTEMHANDLE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_DESTROYSYSTEMHANDLE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_CONNECT,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_DISCONNECT,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_SETUSERDEFAULTVOLUMEID,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETUSERDEFAULTVOLUMEID,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETVOLUMEID,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_MOUNTDB,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_DISMOUNTDB,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_ALLOCHANDLE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_FREEHANDLE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_MOUNT,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_DISMOUNT,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TRANSBEGIN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TRANSCOMMIT,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TRANSABORT,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_PREPARE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_EXECUTE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_EXECDIRECT,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_NEXT,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETDATA,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETMULTICOLUMNDATA,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_PUTDATA,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETOID,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETNUMRESULTCOLS,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETRESULTCOLNAME,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETRESULTCOLTYPE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETRESULTCOLLENGTH,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETERRORMESSAGE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETERRORNAME,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETQUERYERRORMESSAGE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_OIDTOOIDSTRING,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_ADDKEYWORDEXTRACTOR,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_ADDDEFAULTKEYWORDEXTRACTOR,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_DROPKEYWORDEXTRACTOR,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_SETKEYWORDEXTRACTOR,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_ADDFILTER,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_DROPFILTER,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_SETFILTER,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_MAKEINDEX,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_KEYWORDINFOSCAN_OPEN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_KEYWORDINFOSCAN_CLOSE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_KEYWORDINFOSCAN_NEXT,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_KEYWORDINFOSCANFORDOCUMENT_OPEN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_KEYWORDINFOSCANFORDOCUMENT_CLOSE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_KEYWORDINFOSCANFORDOCUMENT_NEXT,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_DEFINEPOSTINGSTRUCTURE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETLOCALTIMEZONE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_SETCURTIME,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETHOUR,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETMINUTE,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETSECOND,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETYEAR,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETMONTH,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETDAY,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETVERSIONSTRING,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETTIMEELAPSED,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_RESETTIMEELAPSED,

	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_CREATESYSTEMHANDLE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_DESTROYSYSTEMHANDLE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_CONNECT_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_DISCONNECT_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_SETUSERDEFAULTVOLUMEID_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETUSERDEFAULTVOLUMEID_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETVOLUMEID_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_MOUNTDB_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_DISMOUNTDB_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_ALLOCHANDLE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_FREEHANDLE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_MOUNT_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_DISMOUNT_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TRANSBEGIN_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TRANSCOMMIT_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TRANSABORT_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_PREPARE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_EXECUTE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_EXECDIRECT_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_NEXT_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETDATA_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETMULTICOLUMNDATA_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_PUTDATA_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETOID_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETNUMRESULTCOLS_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETRESULTCOLNAME_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETRESULTCOLTYPE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETRESULTCOLLENGTH_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETERRORMESSAGE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETERRORNAME_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETQUERYERRORMESSAGE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_OIDTOOIDSTRING_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_ADDKEYWORDEXTRACTOR_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_ADDDEFAULTKEYWORDEXTRACTOR_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_DROPKEYWORDEXTRACTOR_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_SETKEYWORDEXTRACTOR_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_ADDFILTER_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_DROPFILTER_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_SETFILTER_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_MAKEINDEX_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_KEYWORDINFOSCAN_OPEN_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_KEYWORDINFOSCAN_CLOSE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_KEYWORDINFOSCAN_NEXT_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_KEYWORDINFOSCANFORDOCUMENT_OPEN_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_KEYWORDINFOSCANFORDOCUMENT_CLOSE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_KEYWORDINFOSCANFORDOCUMENT_NEXT_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_TEXT_DEFINEPOSTINGSTRUCTURE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETLOCALTIMEZONE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_SETCURTIME_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETHOUR_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETMINUTE_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETSECOND_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETYEAR_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETMONTH_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETDAY_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETVERSIONSTRING_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_GETTIMEELAPSED_RETURN,
	OOSQL_EXTPROCMSG_ID_CALL_OOSQL_RESETTIMEELAPSED_RETURN

};

#endif

